
/*
  Takes the control bits generated by generateControlBits(), a buffer of data to permute,
  and the block size of data elements in the buffer.
  It returns buffer, with the elements permuted in place by the permutation used to 
  create the controlbits.
  
  - control_bits  : the bit array returned by generateControlBits
  - C             : len of control_bits array
  - buffer        : the data buffer to permute
  - block_size    : block size of data elements in buffer 
*/

template <OSwap_Style oswap_style>
void applyPermutation(bool *control_bits, size_t C, unsigned char *buffer, size_t N,
        size_t block_size){
  int m = log2(N);

  for(int i=0; i<2*m-1; i++) {
    size_t gap = 1 << min(i, 2*m-2-i);
    //printf("AP: Layer = %d, gap = %d\n", i, gap);
    for(size_t j=0; j<N/2; j++) {
      size_t pos = (j%gap) + 2*gap*(j/gap);
      bool swap_flag = control_bits[i*(N/2)+j];
      //printf("Swap (%d, %d) ? %d\n", pos, pos + gap, swap_flag);
      oswap_buffer<oswap_style>(buffer+(pos*block_size), 
                      buffer+(pos+gap)*block_size, block_size, swap_flag);
    } 
  }
}
