#!/usr/bin/python3

import sys
import subprocess
import numpy as np
from numpy import genfromtxt
import matplotlib.pyplot as plt
import numpy.polynomial.polynomial as poly
#from scipy.optimize import curve_fit
#from operator import add
#import operator

###############################################################################

# CONFIGS:

# TIME_FLAG :
# 1: plots with pure computation time of the shuffle/sort 
#    (ignores encryption/decryption)
# 2: plots the entire time taken including encryption and decryption
TIME_FLAG=1

# Tell the plotter the result folder
# For Compaction graphs (Fig 11a, 11b, and 14)
RESULTS_FOLDER = "TC_8"
# For Shuffles graph (Fig 12a and Fig 15):
# RESULTS_FOLDER = "Shuffles_8"


# PLOT_MODES:
# 1: Tight Compaction (Goodrich vs ORCompact) OSWAPS (b = 8)
# 2: Tight Compaction (Goodrich vs ORCompact) Time (b = 8)
# 3: Shuffles Time (b = 8)
# 4: Tight Compaction (Goodrich vs ORCompact) Time (b = 8). [Small n version]
# 5: Shuffles Time (b = 8). [Small n version]
PLOT_MODE = 1

###############################################################################

# Auto-select other parameters depending on CONFIGS


'''
MODE MAP:
21 = ORCompact
22 = Goodrich Compact

2 = ORShuffle
3 = BORPStream V1
5 = Bitonic Shuffle
6 = OddEvenMergeShuffle
8 = BORPCompact
33 = BORPStream V2
'''

if(PLOT_MODE == 1 or PLOT_MODE == 2 or PLOT_MODE == 4):
  MODE = [22, 21]
  LABEL_STRINGS = ["Goodrich Compaction", "ORCompact"]
  if(PLOT_MODE == 1):
    save_location = RESULTS_FOLDER + '/TC_OSWAP.png'
  elif(PLOT_MODE == 2):
    save_location = RESULTS_FOLDER + '/TC_time_8.png'
  else:
    save_location = RESULTS_FOLDER + '/TC_time_8_smalln.png'

else:
  MODE = [8, 4, 3, 33, 5, 2]
  LABEL_STRINGS = ["BORPCompact", "OddEvenMerge Shuffle", "BORPStream V1", "BORPStream V2", "Bitonic Shuffle", "ORShuffle"]

  if(PLOT_MODE==3):
    save_location = RESULTS_FOLDER+ '/Shuffles_time_8.png'
  else:
    save_location = RESULTS_FOLDER+ '/Shuffles_time_8_smalln.png'

  #plt.ylim(0,20)
  #plt.xlim(500000,20000000)

# The timing graphs only use item size b = 8
BLOCK_SIZE = [8]

POINT_STYLES = [".","v","s","+","x","d",">","^","*","o","D"]
LINE_STYLES = ['solid', 'dotted', 'dashed', 'dashdot']

###############################################################################

# Plot font size manipulators:

title_font_size = 16
legend_font_size = 12
tick_font_size = 16
axis_label_font_size = 16

###############################################################################

ps_ctr=0
label_ctr=0

# NOTE: Switch the below two lines to disable legend easily
if(PLOT_MODE == 3 or PLOT_MODE == 4 or PLOT_MODE == 5):
  plt.figure(figsize=(8,5))
ax = plt.subplot()
ax.tick_params(axis='both', which='major', labelsize=tick_font_size)
plt.rc('xtick', labelsize=tick_font_size)    # fontsize of the tick labels
plt.rc('ytick', labelsize=tick_font_size)    # fontsize of the tick labels

for m in MODE:
  for b in BLOCK_SIZE: 

    #Gather the data for the requested plot
    # ttime = Total time (including encryption and decryption)
    # ptime = Pure time (exculding encryption and decryption)
    ttime = []
    ptime = []
    ttime_stddev=[]
    ptime_stddev=[]
    OSWAP_count=[]
    lltime=[]
    lltime_stddev=[]
    N = []
  
    log_file_name = RESULTS_FOLDER+"/"+str(m)+"_"+str(b)+".lg" 
    log_file = open(log_file_name, "r")
	
    if(PLOT_MODE==3):
      div = 1000.0
    else:
      div = 1.0

    for line in log_file:
      values = line.split(',')
      N.append(float(values[0].strip()))
      ttime.append(float(values[1].strip())) 
      ttime_stddev.append(float(values[2].strip())) 
      ptime.append(float(values[3].strip())/div) 
      ptime_stddev.append(float(values[4].strip())/div)
      if(m==33 or m ==3):
        lltime.append(float(values[5].strip())/div)
        lltime_stddev.append(float(values[6].strip())/div) 
      OSWAP_count.append(int(values[7].strip()))

    #plt.subplot(1,2,1)
    plt.xlabel('Number of data blocks', fontsize = axis_label_font_size)
    plt.ylabel('Computation time (in s)', fontsize = axis_label_font_size)
    plt.grid(True)

    label_str1 = LABEL_STRINGS[label_ctr] #+", b="+str(b)
    #label_str1 = LABEL_STRINGS[label_ctr]+", b="+str(b)
    #label_str2 = LABEL_STRINGS[m-1]+", b="+str(b)+ " (with enc/dec)" 

    if(TIME_FLAG==1):
      time_to_plot = ptime
      time_to_plot_stddev = ptime_stddev
    else:
      time_to_plot = ttime 
      time_to_plot_stddev = ttime_stddev


    if(PLOT_MODE==1): 
      #Unmute this to plot OSWAPs vs number of data blocks, and mute the subplot (1,2,1) portion
      #plt.subplot(1,2,2)
      plt.xlabel('Number of data blocks')
      plt.ylabel('Number of OSWAPs')
      plt.grid(True)
      points1 = plt.plot(N, OSWAP_count, POINT_STYLES[ps_ctr], label=label_str1, color = str("C")+str(ps_ctr), linestyle='dotted')
    
    else:
      # If mode is 33, we have to show the total time and the phase2 only (ll)time
      # Just plot it with the last color number possible (based on the number of modes we are plotting)
      points1 = plt.plot(N, time_to_plot, POINT_STYLES[ps_ctr], label=label_str1, color = str("C")+str(ps_ctr), linestyle=LINE_STYLES[ps_ctr%len(LINE_STYLES)])
      errbar1 = plt.errorbar(N, time_to_plot, time_to_plot_stddev, None,'r', ls='none', mew=3)
      if(m==33):
        points2 = plt.plot(N, lltime, POINT_STYLES[ps_ctr], label = (LABEL_STRINGS[label_ctr]+" Phase2"), color = str("C")+str(len(MODE)+2), linestyle=LINE_STYLES[ps_ctr%len(LINE_STYLES)])
        #points2 = plt.plot(N, lltime, POINT_STYLES[ps_ctr], label = (LABEL_STRINGS[label_ctr]+" Phase2, b="+str(b)), color = str("C")+str(len(MODE)+2), linestyle=LINE_STYLES[ps_ctr%len(LINE_STYLES)])
        errbar2 = plt.errorbar(N, lltime, lltime_stddev, None,'r', ls='none', mew=3)
      elif(m==3):
        points2 = plt.plot(N, lltime, POINT_STYLES[ps_ctr], label = (LABEL_STRINGS[label_ctr]+" Phase2"), color = str("C")+str(len(MODE)+1), linestyle=LINE_STYLES[ps_ctr%len(LINE_STYLES)])
        errbar2 = plt.errorbar(N, lltime, lltime_stddev, None,'r', ls='none', mew=3)

  ps_ctr+=1
  label_ctr+=1

handles, labels = ax.get_legend_handles_labels()

if(PLOT_MODE==1):
  plt.title('Number of OSWAPS vs Number of datablocks', fontdict = {'fontsize' : title_font_size})
  ax.legend(handles, labels, loc='upper center', bbox_to_anchor=(0.27, 0.95), ncol=1, shadow=True, prop={'size': legend_font_size})
else:
  if(PLOT_MODE==3):
    unit = 's'
  else:
    unit = 'ms'
  plt.title('Computation time (in '+ unit +') vs Number of datablocks', fontdict = {'fontsize' : title_font_size})
  
  if(PLOT_MODE==2 or PLOT_MODE ==4):
    ax.legend(handles, labels, loc='upper center', bbox_to_anchor=(0.27, 0.95), ncol=1, shadow=True, prop={'size': legend_font_size})
  else:
    ax.legend(handles, labels, loc='upper center', bbox_to_anchor=(0.5, -0.15), ncol=4, shadow=True, prop={'size': legend_font_size})


plt.xscale('log', base=2)

# Setting plot size limits:
if(PLOT_MODE==1):
  plt.xlim(100000, 13000000)
  plt.xticks([131072, 262144, 524288, 1048576, 2097152, 4194304, 8388608])
  plt.ylim(0,250000000)
elif(PLOT_MODE==2):
  plt.xlim(100000, 13000000)
  plt.xticks([131072, 262144, 524288, 1048576, 2097152, 4194304, 8388608])
  plt.ylim(0, 1300)
elif(PLOT_MODE==3):
  plt.ylim(0,20)
  plt.xlim(500000, 20000000)
elif(PLOT_MODE==4):
  plt.xticks([16384, 32768, 65536, 131072, 262144])
  plt.ylim(0, 20)
  plt.xlim(10000, 500000)
elif(PLOT_MODE==5):
  plt.xticks([16384, 32768, 65536, 131072, 262144])
  plt.ylim(0,300)
  plt.xlim(10000,500000)

plt.savefig(save_location, bbox_inches='tight')
plt.close()

