#!/usr/bin/python3

import sys
import subprocess
import numpy as np
from numpy import genfromtxt
import matplotlib.pyplot as plt
import numpy.polynomial.polynomial as poly
#from scipy.optimize import curve_fit
#from operator import add
#import operator


###############################################################################

# CONFIGS:

# Tell the plotter the result folder 
# For Compaction graph (Fig 11c)
RESULTS_FOLDER = "TC_blocksizes"
# For Shuffles graph (Fig 12b):
# RESULTS_FOLDER = "Shuffles_blocksizes"

# PLOT_MODES:
# 1: Tight Compaction (Goodrich vs ORCompact) Time
# 2: Shuffles Time
PLOT_MODE = 1

###############################################################################

# Auto-select other parameters depending on CONFIGS

if(PLOT_MODE == 1):
  MODE = [22, 21]
  N = [10000000]
  BLOCK_SIZE_OG = [16, 32, 64, 128, 256, 304, 512, 768, 1024]
  LABEL_STRINGS = ["Goodrich Compaction", "ORCompact"]
  save_location = RESULTS_FOLDER+ '/TC_blocksizes.png'
  plt.ylim(0,40)

else:
  MODE = [8, 4, 3, 33, 5, 2]
  BLOCK_SIZE_OG = [16, 32, 64, 128, 256, 304, 512, 768, 1024, 1280, 1536, 2048, 3072, 4096]
  N = [1048576]
  LABEL_STRINGS = ["BORPCompact", "OddEvenMerge Shuffle", "BORPStream V1", "BORPStream V2", "Bitonic Shuffle", "ORShuffle"]
  save_location = RESULTS_FOLDER+ '/Shuffles_blocksizes.png'

BLOCK_SIZE_M4 = [16, 32, 64, 128, 256, 304, 512, 768, 1024, 1280, 1536, 2048]
POINT_STYLES = [".","v","s","+","x","d",">","^","*","o","D"]
LINE_STYLES = ['solid', 'dotted', 'dashed', 'dashdot']

###############################################################################

# Plot font size manipulators:
title_font_size = 16
legend_font_size = 12
tick_font_size = 16
axis_label_font_size = 16

###############################################################################

ttime = {}
ptime = {}
ttime_stddev={}
ptime_stddev={}
OSWAP_count={}
lltime={}
lltime_stddev={}

label_ctr=0
color_ctr=0
ps_ctr=0
ls_ctr=0

if(PLOT_MODE==2):
  plt.figure(figsize=(8,5))
plt.xlabel('Block size in bytes', fontsize = axis_label_font_size)
plt.ylabel('Computation time (in s)', fontsize = axis_label_font_size)
plt.grid(True)
#plt.yscale('log')
#plt.xscale('log')

for m in MODE:
  for n in N:
    ttime[str(n)]=[]
    ptime[str(n)]=[]
    ttime_stddev[str(n)]=[]
    ptime_stddev[str(n)]=[]
    lltime[str(n)]=[]
    lltime_stddev[str(n)]=[]
    OSWAP_count[str(n)]=[] 

  # If Mode is 4, larger blocksizes go out of graph, and we don't calculate them so we
  # just plot upto where we can see them on the graph
  if(m in {4}):
    BLOCK_SIZE = BLOCK_SIZE_M4
  else:
    BLOCK_SIZE = BLOCK_SIZE_OG
  
  #For each mode we have to get the time for each n in N, for all the block_sizes
  for b in BLOCK_SIZE:   
    log_file_name = RESULTS_FOLDER+"/"+str(m)+"_"+str(b)+".lg" 
    log_file = open(log_file_name, "r")

    for line in log_file:
      values = line.split(',')
      n = int(values[0].strip()) 
      if(n in N):
        key = str(n)
        ttime[key].append(float(values[1].strip()))
        ttime_stddev[key].append(float(values[2].strip()))
        #ptime[key].append(float(values[3].strip())) 
        #ptime_stddev[key].append(float(values[4].strip()))
        #lltime[key].append(float(values[5].strip()))
        #lltime_stddev[key].append(float(values[6].strip()))
        ptime[key].append(float(values[3].strip())/1000.0) 
        ptime_stddev[key].append(float(values[4].strip())/1000.0)
        lltime[key].append(float(values[5].strip())/1000.0)
        lltime_stddev[key].append(float(values[6].strip())/1000.0)
        OSWAP_count[key].append(int(values[7].strip()))

  for n in N:
    print(m, label_ctr)
    if(PLOT_MODE==1):
      label_str1 = LABEL_STRINGS[label_ctr] +", n = 10M"
    else:
      label_str1 = LABEL_STRINGS[label_ctr] #+", N="+str(n)

    points1 = plt.plot(BLOCK_SIZE, ptime[str(n)], POINT_STYLES[ps_ctr], label=label_str1, color = str("C")+str(color_ctr), linestyle=LINE_STYLES[ls_ctr%len(LINE_STYLES)])
    errbar1 = plt.errorbar(BLOCK_SIZE, ptime[str(n)], ptime_stddev[str(n)], None,'r', ls='none', mew=3)
    if(m==33):
      points2 = plt.plot(BLOCK_SIZE, lltime[str(n)], POINT_STYLES[ps_ctr], label = (LABEL_STRINGS[label_ctr]+" Phase2"), color = str("C")+str(len(MODE)+2), linestyle=LINE_STYLES[ps_ctr%len(LINE_STYLES)])
      #points2 = plt.plot(BLOCK_SIZE, lltime[str(n)], POINT_STYLES[ps_ctr], label = (LABEL_STRINGS[label_ctr]+" Phase2, N="+str(n)), color = str("C")+str(len(MODE)+3), linestyle=LINE_STYLES[ps_ctr%len(LINE_STYLES)])
      errbar2 = plt.errorbar(BLOCK_SIZE, lltime[str(n)], lltime_stddev[str(n)], None,'r', ls='none', mew=3)
    elif(m==3):
      points2 = plt.plot(BLOCK_SIZE, lltime[str(n)], POINT_STYLES[ps_ctr], label = (LABEL_STRINGS[label_ctr]+" Phase2"), color = str("C")+str(len(MODE)+1), linestyle=LINE_STYLES[ps_ctr%len(LINE_STYLES)])
      #points2 = plt.plot(BLOCK_SIZE, lltime[str(n)], POINT_STYLES[ps_ctr], label = (LABEL_STRINGS[label_ctr]+" Phase2, b="+str(b)), color = str("C")+str(len(MODE)+1), linestyle=LINE_STYLES[ps_ctr%len(LINE_STYLES)])
      errbar2 = plt.errorbar(BLOCK_SIZE, lltime[str(n)], lltime_stddev[str(n)], None,'r', ls='none', mew=3)

    color_ctr+=1
    ps_ctr+=1
     
  label_ctr+=1
  ls_ctr+=1

ax = plt.subplot()
ax.tick_params(axis='both', which='major', labelsize=tick_font_size)
handles, labels = ax.get_legend_handles_labels()

if(PLOT_MODE==1):
  #ax.set_xticks([16, 32, 64, 128, 256, 304, 512, 768, 1024])
  #ax.set_xticklabels([r'$2^4$', r'$2^5$', r'$2^6$', r'$2^7$', r'$2^8$', r'$2^{8.5}$', r'$2^9$', r'$2^{9.5}$', r'$2^{10}$'])
  ax.set_xticks([16, 128, 256, 512, 768, 1024])
  ax.set_xticklabels([r'$2^4$', r'$2^7$', r'$2^8$', r'$2^9$', r'$2^{9.5}$', r'$2^{10}$'])
else:
  plt.xscale('log', base=2)
  #plt.set_xticks()
  plt.xlim(64, 5000)
  plt.ylim(0,550)

#plt.title('Computation time (in ms) vs block size in bytes', fontdict = {'fontsize' : title_font_size} )
plt.title('Computation time (in s) vs block size in bytes', fontdict = {'fontsize' : title_font_size} )
if(PLOT_MODE==1):
  ax.legend(handles, labels, loc='upper center', bbox_to_anchor=(0.35, 0.95), ncol=1, shadow=True, prop={'size': legend_font_size})
else:
  ax.legend(handles, labels, loc='upper center', bbox_to_anchor=(0.5, -0.15), ncol=4, shadow=True, prop={'size': legend_font_size})

plt.savefig(save_location, bbox_inches='tight')
plt.close()
